/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 
*/

#include <CCore>

#include "RBoard.h"
#include "RFBoard.h"

inline void cRFastBoard::
StuffRow(LongLine rLine, Position row)
{
	this->Board[row] = rLine;

#define ROWSTFDXY(i) this->BoardXY[row+(i)] =  cReversiBoardTransferPiece(rLine, this->BoardXY[row+(i)], (7-(i)))
#define ROWSTFDYX(i) this->BoardYX[7+row-(i)] = cReversiBoardTransferPiece(rLine, this->BoardYX[7+row-(i)], (7-(i)))

	ROWSTFDXY(0);ROWSTFDXY(1);ROWSTFDXY(2);ROWSTFDXY(3);ROWSTFDXY(4);ROWSTFDXY(5);ROWSTFDXY(6);ROWSTFDXY(7);
	ROWSTFDYX(0);ROWSTFDYX(1);ROWSTFDYX(2);ROWSTFDYX(3);ROWSTFDYX(4);ROWSTFDYX(5);ROWSTFDYX(6);ROWSTFDYX(7);
#undef ROWSTFDXY
#undef ROWSTFDYX

#define VERTICALSTUFF(i)\
	this->BoardY[i] = cReversiBoardTransferPiece(rLine, this->BoardY[i], (7-(row)));\
	rLine >>=kPieceBits;

	rLine <<= kPieceBits*(7-row);

	VERTICALSTUFF(7);
	VERTICALSTUFF(6);
	VERTICALSTUFF(5);
	VERTICALSTUFF(4);
	VERTICALSTUFF(3);
	VERTICALSTUFF(2);
	VERTICALSTUFF(1);
	VERTICALSTUFF(0);

#undef VERTICALSTUFF
}

inline void cRFastBoard::
StuffColumn(LongLine rLine, Position column)
{
	this->BoardY[column] = rLine;

	rLine <<= kPieceBits*(7-column);

#define CSTFD(i)\
		this->Board[i] = cReversiBoardTransferPiece(rLine, this->Board[i], 7-(column));\
		this->BoardXY[column+(i)] = cReversiBoardTransferPiece(rLine, this->BoardXY[column+(i)], 7-(column));\
		this->BoardYX[7-column+(i)] = cReversiBoardTransferPiece(rLine, this->BoardYX[7-column+(i)], 7-(column));\
		rLine >>= kPieceBits;

	CSTFD(7);
	CSTFD(6);
	CSTFD(5);
	CSTFD(4);
	CSTFD(3);
	CSTFD(2);
	CSTFD(1);
	CSTFD(0);

#undef CSTFD
}

void cRFastBoard::
StuffXY(LongLine rLine, Position xy)
{
#define DSXY(i,sh)\
			this->Board[i] = cReversiBoardTransferPiece(rLine, this->Board[i], (sh));\
			this->BoardY[7-(sh)] = cReversiBoardTransferPiece(rLine, this->BoardY[7-(sh)], (7-(i)));\

#define LDSXY(i,sh,lsh)\
			this->Board[i] = cReversiBoardTransferPiece(rLine, this->Board[i], (sh));\
			this->BoardY[7-(sh)] = cReversiBoardTransferPiece((rLine << ((lsh)*kPieceBits)),this->BoardY[7-(sh)], (7-(i)));

#define RDSXY(i,sh,rsh)\
			this->Board[i] = cReversiBoardTransferPiece(rLine, this->Board[i], (sh));\
			this->BoardY[7-(sh)] = cReversiBoardTransferPiece((rLine >> ((rsh)*kPieceBits)),this->BoardY[7-(sh)], (7-(i)));

#define DSXY1(d,s)	RDSXY(d,s,1);LDSXY(d-1,s-1,1)
#define DSXY2(d,s)	RDSXY(d,s,2);DSXY(d-1,s-1);LDSXY(d-2,s-2,2)
#define DSXY3(d,s)	RDSXY(d,s,3);DSXY1(d-1,s-1);LDSXY(d-3,s-3,3)
#define DSXY4(d,s)	RDSXY(d,s,4);DSXY2(d-1,s-1);LDSXY(d-4,s-4,4)
#define DSXY5(d,s)	RDSXY(d,s,5);DSXY3(d-1,s-1);LDSXY(d-5,s-5,5)
#define DSXY6(d,s)	RDSXY(d,s,6);DSXY4(d-1,s-1);LDSXY(d-6,s-6,6)
#define DSXY7(d,s)	RDSXY(d,s,7);DSXY5(d-1,s-1);LDSXY(d-7,s-7,7)

#define xySYX(d,i) this->BoardYX[7+(d)] = cReversiBoardTransferPiece(rLine, this->BoardYX[7+(d)], (i))
#define xySYX1(d,i) xySYX(d,i); xySYX((d)-2,(i)-1)
#define xySYX2(d,i) xySYX(d,i); xySYX1((d)-2,(i)-1)
#define xySYX3(d,i) xySYX(d,i); xySYX2((d)-2,(i)-1)
#define xySYX4(d,i) xySYX(d,i); xySYX3((d)-2,(i)-1)
#define xySYX5(d,i) xySYX(d,i); xySYX4((d)-2,(i)-1)
#define xySYX6(d,i) xySYX(d,i); xySYX5((d)-2,(i)-1)
#define xySYX7(d,i) xySYX(d,i); xySYX6((d)-2,(i)-1)

	this->BoardXY[xy] = rLine;

	switch(xy)
	{
		case 0:		DSXY(0,7);	xySYX(0,7); break;
		case 1:		DSXY1(1,7);	xySYX1(1,7); break;
		case 2:		DSXY2(2,7);	xySYX2(2,7); break;
		case 3:		DSXY3(3,7);	xySYX3(3,7); break;
		case 4:		DSXY4(4,7);	xySYX4(4,7); break;
		case 5:		DSXY5(5,7);	xySYX5(5,7); break;
		case 6:		DSXY6(6,7);	xySYX6(6,7); break;
		case 7:		DSXY7(7,7);	xySYX7(7,7); break;
		case 8:		DSXY6(7,6); xySYX6(6,6); break;
		case 9:		DSXY5(7,5); xySYX5(5,5); break;
		case 10:	DSXY4(7,4); xySYX4(4,4); break;
		case 11:	DSXY3(7,3); xySYX3(3,3); break;
		case 12:	DSXY2(7,2); xySYX2(2,2); break;
		case 13:	DSXY1(7,1); xySYX1(1,1); break;
		case 14:	DSXY(7,0);	xySYX(0,0); break;
	}


#undef xySYX7
#undef xySYX6
#undef xySYX5
#undef xySYX4
#undef xySYX3
#undef xySYX2
#undef xySYX1
#undef xySYX

#undef DSXY7
#undef DSXY6
#undef DSXY5
#undef DSXY4
#undef DSXY3
#undef DSXY2
#undef DSXY1
#undef LDSXY
#undef RDSXY
#undef DSXY
}


void cRFastBoard::
StuffYX(LongLine rLine, Position yx)
{
#define LDSYX(i,sh)\
		this->Board[i] = cReversiBoardTransferPiece(rLine, this->Board[i], (sh));\
		this->BoardY[7-(sh)] = cReversiBoardTransferPiece(rLine << ((7-(sh)-(i))*kPieceBits),this->BoardY[7-(sh)], (7-(i)));

#define RDSYX(i,sh)\
		this->Board[i] = cReversiBoardTransferPiece(rLine, this->Board[i], (sh));\
		this->BoardY[7-(sh)] = cReversiBoardTransferPiece(rLine >> (((sh)+(i)-7)*kPieceBits),this->BoardY[7-(sh)], (7-(i)));

#define LDSYX1(d,s)	LDSYX(d,s)
#define LDSYX2(d,s)	LDSYX1(d,s);LDSYX(d-1,s+1)
#define LDSYX3(d,s)	LDSYX2(d,s);LDSYX(d-2,s+2)
#define LDSYX4(d,s)	LDSYX3(d,s);LDSYX(d-3,s+3)
#define LDSYX5(d,s)	LDSYX4(d,s);LDSYX(d-4,s+4)
#define LDSYX6(d,s)	LDSYX5(d,s);LDSYX(d-5,s+5)
#define LDSYX7(d,s)	LDSYX6(d,s);LDSYX(d-6,s+6)
#define LDSYX8(d,s)	LDSYX7(d,s);LDSYX(d-7,s+7)

#define RDSYX1(d,s)	RDSYX(d,s)
#define RDSYX2(d,s)	RDSYX1(d,s);RDSYX(d-1,s+1)
#define RDSYX3(d,s)	RDSYX2(d,s);RDSYX(d-2,s+2)
#define RDSYX4(d,s)	RDSYX3(d,s);RDSYX(d-3,s+3)
#define RDSYX5(d,s)	RDSYX4(d,s);RDSYX(d-4,s+4)
#define RDSYX6(d,s)	RDSYX5(d,s);RDSYX(d-5,s+5)
#define RDSYX7(d,s)	RDSYX6(d,s);RDSYX(d-6,s+6)

#define yxSXY(d,i) this->BoardXY[7-(d)] = cReversiBoardTransferPiece(rLine, this->BoardXY[7-(d)], (i))
#define yxSXY1(d,i) yxSXY(d,i); yxSXY((d)-2,(i)-1)
#define yxSXY2(d,i) yxSXY(d,i); yxSXY1((d)-2,(i)-1)
#define yxSXY3(d,i) yxSXY(d,i); yxSXY2((d)-2,(i)-1)
#define yxSXY4(d,i) yxSXY(d,i); yxSXY3((d)-2,(i)-1)
#define yxSXY5(d,i) yxSXY(d,i); yxSXY4((d)-2,(i)-1)
#define yxSXY6(d,i) yxSXY(d,i); yxSXY5((d)-2,(i)-1)
#define yxSXY7(d,i) yxSXY(d,i); yxSXY6((d)-2,(i)-1)

	this->BoardYX[yx] = rLine;

	switch(yx)
	{
		case 0:		LDSYX1(0,0);	yxSXY(0,0);		break;
		case 1:		LDSYX2(1,0);	yxSXY1(1,1);	break;
		case 2:		LDSYX3(2,0);	yxSXY2(2,2);	break;
		case 3:		LDSYX4(3,0);	yxSXY3(3,3);	break;
		case 4:		LDSYX5(4,0);	yxSXY4(4,4);	break;
		case 5:		LDSYX6(5,0);	yxSXY5(5,5);	break;
		case 6:		LDSYX7(6,0);	yxSXY6(6,6);	break;
		case 7:		LDSYX8(7,0);	yxSXY7(7,7);	break;
		case 8:		RDSYX7(7,1);	yxSXY6(6,7);	break;
		case 9:		RDSYX6(7,2);	yxSXY5(5,7);	break;
		case 10:	RDSYX5(7,3);	yxSXY4(4,7);	break;
		case 11:	RDSYX4(7,4);	yxSXY3(3,7);	break;
		case 12:	RDSYX3(7,5);	yxSXY2(2,7);	break;
		case 13:	RDSYX2(7,6);	yxSXY1(1,7);	break;
		case 14:	RDSYX1(7,7);	yxSXY(0,7);		break;
	}


#undef yxSXY7
#undef yxSXY6
#undef yxSXY5
#undef yxSXY4
#undef yxSXY3
#undef yxSXY2
#undef yxSXY1
#undef yxSXY

#undef RDSYX7
#undef RDSYX6
#undef RDSYX5
#undef RDSYX4
#undef RDSYX3
#undef RDSYX2
#undef RDSYX1
#undef RDSYX

#undef LDSYX8
#undef LDSYX7
#undef LDSYX6
#undef LDSYX5
#undef LDSYX4
#undef LDSYX3
#undef LDSYX2
#undef LDSYX1
#undef LDSYX
}

inline void cRFastBoard::
SetPiece(Piece p, Position x, Position y)
{
	static const LongLine rgAllPieces[]={0,0,kAllPiece1, kAllPiece2};

	this->BoardY[x] = cReversiBoardTransferPiece(rgAllPieces[p], this->BoardY[x], 7-y);
	this->Board[y] = cReversiBoardTransferPiece(rgAllPieces[p], this->Board[y], 7-x);
	this->BoardXY[x+y] = cReversiBoardTransferPiece(rgAllPieces[p], this->BoardXY[x+y], 7-x);
	this->BoardYX[y+7-x] = cReversiBoardTransferPiece(rgAllPieces[p], this->BoardYX[y+7-x], 7-x);
}

void cRFastBoard::
MakeMove(Piece p, Position x, Position y)
{
	const ulong rlPieceNormalizerMask = cRFastBoard::grgPieceNormalizerMask[p];

	Assert((p==kPiece1)||(p==kPiece2));
	Assert(x<8 && y<8);

#define CAPTURELINE(rg, stuffproc, i, pos)\
	if(cReversiBoard::IsCapturePosition(this->##rg[i] ^ rlPieceNormalizerMask, pos))\
		this->##stuffproc(cRFastBoard::CaptureLine(this->##rg[i] ^ rlPieceNormalizerMask, pos) ^ rlPieceNormalizerMask, i);
		
	CAPTURELINE(BoardYX, StuffYX, 7+y-x, x);
	CAPTURELINE(BoardXY, StuffXY, y+x, x);
	CAPTURELINE(Board, StuffRow, y, x);
	CAPTURELINE(BoardY, StuffColumn, x, y);

#undef CAPTURELINE
	this->SetPiece(p, x, y);

//	this->UpdateMobility(cRFastBoard::OtherPiece(p));

#ifdef DEBUG
	//this->CrosscheckBoard();
#endif
}

#ifdef DEBUG
void cRFastBoard::
CrosscheckBoard()
{
	unsigned i;
	LongLine rLine;

	for (i=0; i<8; i++)
		Assert(this->BoardY[i] == (rLine = this->ExtractY(i)));

	for (i=0; i<15; i++)
		Assert(this->BoardXY[i] == (rLine=this->ExtractXY(i)));

	for (i=0; i<15; i++)
		Assert(this->BoardYX[i] == (rLine=this->ExtractYX(i)));
}

ulong cRFastBoard::
FullLineCount()
{
	unsigned i;
	ulong count=0;

	for (i=0; i<8; i++)
	{
		if(cReversiBoard::IsLineFull(this->BoardY[i]))
			count++;

		if (cReversiBoard::IsLineFull(this->Board[i]))
			count++;
	}

	for (i=0; i<15; i++)
	{
		if (cReversiBoard::IsLineFull(this->BoardXY[i]))
			count++;
		if (cReversiBoard::IsLineFull(this->BoardYX[i]))
			count++;
	}
	return count;
}

ulong cRFastBoard::
EmptyLineCount()
{
	unsigned i;
	ulong count=0;

	for (i=0; i<8; i++)
	{
		if(cReversiBoard::IsLineEmpty(this->BoardY[i]))
			count++;

		if (cReversiBoard::IsLineEmpty(this->Board[i]))
			count++;
	}

	for (i=0; i<15; i++)
	{
		if (cReversiBoard::IsLineEmpty(this->BoardXY[i]))
			count++;
		if (cReversiBoard::IsLineEmpty(this->BoardYX[i]))
			count++;
	}
	return count;
}

#endif
